"""Summarize status of secret files."""

from .encryption import EncryptedConfigFile
import glob
import os.path


class Environment(object):

    def __init__(self, path=None, name=None):
        if path:
            self.path = path
            name = os.path.basename(path)
            self.name = os.path.splitext(name)[0]
        else:
            self.name = name
            self.path = 'secrets/{}.cfg'.format(name)

        self.f = EncryptedConfigFile(self.path, write_lock=True, quiet=True)
        self.f.__enter__()

    @classmethod
    def all(cls):
        for e in glob.glob('secrets/*.cfg'):
            yield Environment(path=e)

    @classmethod
    def by_filter(cls, filter):
        if filter:
            filter = filter.split(',')
        for e in cls.all():
            if filter and e.name not in filter:
                continue
            yield e

    def summary(self):
        try:
            self.f.read()
        except Exception:
            print "\t<Can not decrypt. You seem to not have access.>"
            return
        print "\t", self.f.config.get('batou', 'members')

    def add_user(self, keyid):
        try:
            self.f.read()
        except Exception:
            print "\t<Can not decrypt. You seem to not have access.>"
            return
        members = self.f.get_members()
        if keyid not in members:
            members.append(keyid)
            self.f.set_members(members)
            try:
                self.f.write_config()
            except Exception:
                print "\t<Can not encrypt. You seem to be missing a key.>"
                return
        self.summary()

    def remove_user(self, keyid):
        try:
            self.f.read()
        except Exception:
            print "\t<Can not decrypt. You seem to not have access.>"
            return
        members = self.f.get_members()
        if keyid in members:
            members.remove(keyid)
            self.f.set_members(members)
            try:
                self.f.write_config()
            except Exception:
                print "\t<Can not encrypt. You seem to be missing a key.>"
                return
        self.summary()


def summary(**kw):
    """Secrets editor console script.

    The main focus here is to avoid having unencrypted files accidentally
    ending up in the deployment repository.

    """
    if not os.path.exists('secrets'):
        print "No secrets."

    for e in Environment.all():
        print e.name
        e.summary()
        print


def add_user(keyid, environments, **kw):
    """Add a user to a given environment.

    If environments is not given, a user is added
    to all environments.

    """
    if not os.path.exists('secrets'):
        print "No secrets."

    for e in Environment.by_filter(environments):
        print e.name
        e.add_user(keyid)
        print


def remove_user(keyid, environments, **kw):
    """Remove a user from a given environment.

    If environments is not given, the user is removedd
    from all environments.

    """
    if not os.path.exists('secrets'):
        print "No secrets."

    for e in Environment.by_filter(environments):
        print e.name
        e.remove_user(keyid)
        print



from batou import UpdateNeeded
from batou.component import Component
import os
import tempfile


class Command(Component):

    namevar = 'statement'
    admin_password = ''
    admin_user = 'root'
    hostname = None
    port = None

    db = 'mysql'

    # Unless allows you to specify a query that generates output if the
    # target state has been reached already.
    unless = ''

    def configure(self):
        self.statement = self.expand(self.statement)
        self.unless = self.expand(self.unless)

    def _mysql(self, cmd):
        _, self.tmp = tempfile.mkstemp(suffix='sql')
        with open(self.tmp, 'w') as f:
            f.write(cmd+'\n')

        command = [
            'mysql -Bs -u{{component.admin_user}}',
            '-p{{component.admin_password}}']
        if self.hostname:
            command.append('-h {{component.hostname}}')
        if self.port:
            command.append('-P {{component.port}}')
        command.append('{{component.db}} < {{component.tmp}}')

        out, err = self.cmd(self.expand(' '.join(command)))
        os.remove(self.tmp)
        return out, err

    def verify(self):
        if not self.unless:
            raise UpdateNeeded()
        out, err = self._mysql(self.unless)
        if not out.strip():
            raise UpdateNeeded()

    def update(self):
        self._mysql(self.statement)

    @property
    def namevar_for_breadcrumb(self):
        words = self.statement.split()
        if words:
            return ' '.join(words[:2]).upper()
        return '--missing--'


class Database(Component):

    namevar = 'database'
    charset = 'UTF8'
    base_import_file = None
    admin_password = None

    def configure(self):
        create_db = self.expand("""\
CREATE DATABASE IF NOT EXISTS
    {{component.database}}
    DEFAULT CHARACTER SET = '{{component.charset}}';
""")
        self += Command(create_db, admin_password=self.admin_password)

        if self.base_import_file:
            self += Command(
                self.expand('\. {{component.base_import_file}}\n'),
                db=self.database,
                unless=self.expand('show tables'),
                admin_password=self.admin_password)


class User(Component):

    namevar = 'user'
    password = None
    allow_from_hostname = 'localhost'
    admin_password = None

    def configure(self):

        create = self.expand("""\
CREATE USER '{{component.user}}'@'{{component.allow_from_hostname}}';
""")
        create_unless = self.expand("""\
SELECT *
FROM user
WHERE
    User = '{{component.user}}'
    AND
    Host = '{{component.allow_from_hostname}}';
""")
        self += Command(
            create, unless=create_unless, admin_password=self.admin_password)

        set_password = self.expand("""\
SET PASSWORD FOR
    '{{component.user}}'@'{{component.allow_from_hostname}}' =
     PASSWORD('{{component.password}}');
""")
        self += Command(set_password, admin_password=self.admin_password)


class Grant(Command):

    namevar = 'grant_db'
    user = ''
    allow_from_hostname = 'localhost'
    statement = """\
GRANT ALL
    ON {{component.grant_db}}
    TO '{{component.user}}'@'{{component.allow_from_hostname}}';
"""



from ..host import Host
import mock
import pytest


@pytest.fixture
def host():
    return Host('foo.testing', mock.sentinel.env)


def test_data_can_set_and_get(host):
    # Okay, this test is almost too silly to have it here.
    host.data['foo'] = 'bar'
    assert 'bar' == host.data['foo']



from ..download import Download
import mock
import os.path
import pytest
import unittest


class DownloadTest(unittest.TestCase):

    def test_verify_should_pass_checkum_function_to_hash(self):
        import batou
        component = Download('url', checksum='foobar:1234')
        component.configure()
        with mock.patch('os.path.exists') as exists,\
                mock.patch('batou.utils.hash') as buh:
            exists.return_value = True
            try:
                component.verify()
            except batou.UpdateNeeded:
                pass
        buh.assert_called_with(mock.ANY, 'foobar')

    def test_configure_should_raise_valueerror_if_no_checksum_given(self):
        download = Download('url')
        with self.assertRaises(ValueError) as exc:
            download.configure()
        self.assertEqual('No checksum given.', exc.exception.args[0])

    def test_update_should_raise_AssertionError_on_checksum_mismatch(self):
        download = Download('url', checksum='foobar:1234')
        download.configure()
        with mock.patch('batou.lib.download.urlretrieve') as retrieve, \
                mock.patch('batou.utils.hash') as buh,\
                self.assertRaises(AssertionError) as err:
            retrieve.return_value = 'url', []
            buh.return_value = '4321'
            download.update()
        self.assertEqual('Checksum mismatch!\nexpected: 1234\ngot: 4321',
                         str(err.exception))

    def test_breadcrumbs_should_not_show_password(self):
        component = Download('http://me:mypw@myhost/file.name',
                             checksum='foobar:1234')
        assert 'http://me:*****@myhost/file.name' == \
            component.namevar_for_breadcrumb


@pytest.mark.slow
def test_downloads_file(root):
    root.component += Download(
        'http://speedtest.ftp.otenet.gr/files/test100k.db',
        checksum='md5:4c6426ac7ef186464ecbb0d81cbfcb1e')
    root.component.deploy()
    assert os.path.isfile(os.path.join(
        root.environment.workdir_base, 'mycomponent/test100k.db'))



from batou import remote_core
import inspect
import mock
import os.path
import pytest


@pytest.yield_fixture(autouse=True)
def reset_global_vars():
    yield
    remote_core.target_directory = None
    remote_core.environment = None
    remote_core.deployment = None
    remote_core.channel = None


def test_deployment_and_channel_exist_as_names():
    assert remote_core.channel is None
    assert remote_core.deployment is None


def test_lock():
    remote_core.lock()


def test_cmd():
    result = remote_core.cmd('echo "asdf"')
    assert result == ('asdf\n', '')


@pytest.fixture
def mock_remote_core(monkeypatch):
    # Suppress active actions in the remote_core module
    monkeypatch.setattr(remote_core, 'cmd', mock.Mock())


def test_update_code_existing_target(mock_remote_core, tmpdir):
    remote_core.cmd.side_effect = [
        ('', ''),
        ('', ''),
        ('', ''),
        ('', '')]

    remote_core.ensure_repository(str(tmpdir), 'hg-pull')
    remote_core.hg_pull_code('http://bitbucket.org/flyingcircus/batou')
    remote_core.hg_update_working_copy('default')

    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert calls.next().startswith('hg init /')
    assert calls.next() == 'hg pull http://bitbucket.org/flyingcircus/batou'
    assert calls.next() == 'hg up -C default'
    assert calls.next() == 'hg id -i'
    assert remote_core.cmd.call_count == 4


def test_update_code_new_target(mock_remote_core, tmpdir):
    remote_core.cmd.side_effect = [
        ('', ''),
        ('', ''),
        ('', ''),
        ('', '')]
    remote_core.ensure_repository(str(tmpdir) + '/foo', 'hg-bundle')
    remote_core.hg_pull_code('http://bitbucket.org/flyingcircus/batou')
    remote_core.hg_update_working_copy('default')

    assert remote_core.cmd.call_count == 4
    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert calls.next() == 'hg init {}'.format(remote_core.target_directory)
    assert calls.next() == 'hg pull http://bitbucket.org/flyingcircus/batou'
    assert calls.next() == 'hg up -C default'
    assert calls.next() == 'hg id -i'


def test_hg_bundle_shipping(mock_remote_core, tmpdir):
    remote_core.ensure_repository(str(tmpdir) + '/foo', 'hg-bundle')
    remote_core.cmd.side_effect = [
        ("41fea38ce5d3", None),
        ("""\
changeset: 371:revision-a
summary:fdsa

changeset: 372:revision-b
""", None),
        ('', ''),
        ('', ''),
        ('', '')]
    heads = remote_core.hg_current_heads()
    assert heads == ['revision-a', 'revision-b']
    remote_core.hg_unbundle_code()
    remote_core.hg_update_working_copy('default')

    assert remote_core.cmd.call_count == 6
    calls = iter(x[1][0] for x in remote_core.cmd.mock_calls)
    assert calls.next() == 'hg init {}'.format(remote_core.target_directory)
    assert calls.next() == 'hg id -i'
    assert calls.next() == 'hg heads'
    assert calls.next() == 'hg -y unbundle batou-bundle.hg'
    assert calls.next() == 'hg up -C default'
    assert calls.next() == 'hg id -i'


def test_build_batou_fresh_install(mock_remote_core, tmpdir):
    remote_core.ensure_repository(str(tmpdir), 'hg-pull')
    remote_core.cmd.reset_mock()
    remote_core.build_batou('.', 'asdf')
    calls = iter([x[1][0] for x in remote_core.cmd.mock_calls])
    assert remote_core.cmd.call_count == 1
    assert calls.next() == './batou --help'


def test_build_batou_virtualenv_exists(mock_remote_core, tmpdir):
    remote_core.ensure_repository(str(tmpdir), 'hg-pull')
    os.mkdir(remote_core.target_directory + '/bin')
    open(remote_core.target_directory + '/bin/python2.7', 'w')
    remote_core.build_batou('.', 'asdf')
    calls = iter([x[1][0] for x in remote_core.cmd.mock_calls])
    assert remote_core.cmd.call_count == 2
    calls.next()  # skip ensure_repository
    assert calls.next() == './batou --help'


def test_expand_deployment_base(tmpdir):
    with mock.patch('os.path.expanduser') as expanduser:
        expanduser.return_value = str(tmpdir)
        remote_core.ensure_repository('~/deployment', 'rsync')
    assert (remote_core.target_directory == str(tmpdir))


def test_deploy_component(monkeypatch):
    import batou
    monkeypatch.setattr(remote_core, 'deployment', mock.Mock())
    monkeypatch.setattr(batou, 'output', mock.Mock())
    remote_core.deploy('foo')
    assert remote_core.deployment.deploy.call_count == 1


def test_whoami():
    # Smoke test
    assert remote_core.whoami() != ''


class DummyChannel(object):

    _closed = False

    def __init__(self):
        self.receivequeue = []
        self.sendqueue = []

    def isclosed(self):
        return self._closed

    def receive(self, timeout=None):
        result = self.receivequeue.pop(0)
        if not self.receivequeue:
            self._closed = True
        return result

    def send(self, item):
        self.sendqueue.append(item)


@pytest.fixture
def remote_core_mod():
    channel = DummyChannel()

    local_namespace = dict(channel=channel, __name__='__channelexec__')

    remote_core_mod = compile(
        inspect.getsource(remote_core),
        inspect.getsourcefile(remote_core),
        'exec')

    def run():
        exec remote_core_mod in local_namespace

    return (channel, run)


def test_channelexec_already_closed(remote_core_mod):
    channel, run = remote_core_mod
    channel._closed = True
    run()
    assert channel.receivequeue == []
    assert channel.sendqueue == []


def test_channelexec_echo_cmd(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(('cmd', ('echo "asdf"',), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    assert channel.sendqueue == [('batou-result', ('asdf\n', ''))]


def test_channelexec_multiple_echo_cmds(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(('cmd', ('echo "asdf1"',), {}))
    channel.receivequeue.append(('cmd', ('echo "asdf2"',), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    assert channel.sendqueue == [('batou-result', ('asdf1\n', '')),
                                 ('batou-result', ('asdf2\n', ''))]


def test_channelexec_handle_exception(remote_core_mod):
    channel, run = remote_core_mod
    channel.receivequeue.append(('cmd', ('fdjkahfkjdasbfda',), {}))
    run()
    assert channel.isclosed()
    assert channel.receivequeue == []
    response = channel.sendqueue
    assert response == [
        ('batou-output',
         'line',
         ('ERROR: fdjkahfkjdasbfda',),
         {'bold': True, 'red': True}),
        ('batou-output', 'line', ('     Return code: 127',), {'red': True}),
        ('batou-output', 'line', ('STDOUT',), {'red': True}),
        ('batou-output', 'line', ('     ',), {}),
        ('batou-output', 'line', ('STDERR',), {'red': True}),
        ('batou-output',
         'line',
         ('     /bin/sh: fdjkahfkjdasbfda: command not found\n     ',),
         {}),
        ('batou-error', None)]



from __future__ import print_function, unicode_literals
import ConfigParser
import StringIO
import fcntl
import os
import subprocess
import tempfile

# https://thraxil.org/users/anders/posts/2008/03/13/Subprocess-Hanging-PIPE-is-your-enemy/
NULL = tempfile.TemporaryFile()


NEW_FILE_TEMPLATE = """\
[batou]
members =
"""


class EncryptedConfigFile(object):
    """Wrap encrypted config files."""

    lockfd = None
    _cleartext = None

    # Additional GPG parameters. Used for testing.
    gpg_opts = ''

    def __init__(self, encrypted_file, write_lock=False, quiet=False):
        """Context manager that opens an encrypted file.

        Use the read() and write() methods in the subordinate "with"
        block to manipulate cleartext content. If the cleartext content
        has been replaced, the encrypted file is updated.

        `write_lock` must be set True if a modification of the file is
        intended.
        """
        self.encrypted_file = encrypted_file
        self.write_lock = write_lock
        self.quiet = quiet

    def __enter__(self):
        self._lock()
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.lockfd.close()

    def gpg(self, cmdline):
        null = tempfile.TemporaryFile()
        for gpg in ['gpg2', 'gpg']:
            try:
                subprocess.check_call(
                    [gpg, '--version'],
                    stdout=null, stderr=null)
            except (subprocess.CalledProcessError, OSError):
                pass
            else:
                return '{} {}'.format(gpg, cmdline)

    @property
    def cleartext(self):
        if self._cleartext is None:
            return NEW_FILE_TEMPLATE
        return self._cleartext

    @cleartext.setter
    def cleartext(self, value):
        self.config = ConfigParser.ConfigParser()
        self.config.readfp(StringIO.StringIO(value))
        self.set_members(self.get_members())
        s = StringIO.StringIO()
        self.config.write(s)
        self._cleartext = s.getvalue()

    def read(self):
        if self._cleartext is None:
            if os.stat(self.encrypted_file).st_size:
                self._decrypt()
        return self.cleartext

    def write(self, cleartext):
        """Replace encrypted file with new content."""
        if not self.write_lock:
            raise RuntimeError('write() needs a write lock')
        self.cleartext = cleartext
        self._encrypt()

    def write_config(self):
        s = StringIO.StringIO()
        self.config.write(s)
        self.write(s.getvalue())

    def _lock(self):
        self.lockfd = open(
            self.encrypted_file, self.write_lock and 'a+' or 'r+')
        if self.write_lock:
            fcntl.lockf(self.lockfd, fcntl.LOCK_EX)
        else:
            fcntl.lockf(self.lockfd, fcntl.LOCK_SH)

    def _decrypt(self):
        opts = self.gpg_opts
        if self.quiet:
            opts += ' -q --no-tty --batch'
        self.cleartext = subprocess.check_output(
            [self.gpg('{} --decrypt {}'.format(
                opts, self.encrypted_file))],
            stderr=NULL,
            shell=True)

    def get_members(self):
        members = self.config.get('batou', 'members').split(',')
        members = [x.strip() for x in members]
        members = filter(None, members)
        members.sort()
        return members

    def set_members(self, members):
        # The whitespace here is exactly what
        # "members = " looks like in the config file so we get
        # proper indentation.
        members = ',\n      '.join(members)
        self.config.set('batou', 'members', members)

    def _encrypt(self):
        recipients = self.get_members()
        if not recipients:
            raise ValueError("Need at least one recipient.")
        self.set_members(self.get_members())
        recipients = ' '.join(['-r {}'.format(r.strip()) for r in recipients])
        os.rename(self.encrypted_file,
                  self.encrypted_file + '.old')
        try:
            gpg = subprocess.Popen(
                [self.gpg('{} --encrypt {} -o {}'.format(
                    self.gpg_opts, recipients, self.encrypted_file))],
                stdin=subprocess.PIPE,
                shell=True)
            gpg.communicate(self.cleartext)
            if gpg.returncode != 0:
                raise RuntimeError('GPG returned non-zero exit code.')
        except Exception:
            os.rename(self.encrypted_file + '.old',
                      self.encrypted_file)
            raise
        else:
            os.unlink(self.encrypted_file + '.old')
